"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var NestedSetNode = /** @class */ (function () {
    function NestedSetNode(title, left, right, depth, parent) {
        if (title === void 0) { title = ''; }
        if (left === void 0) { left = 0; }
        if (right === void 0) { right = 1; }
        if (depth === void 0) { depth = 0; }
        this.title = title;
        this.depth = depth;
        this.left = left;
        this.right = right;
        this.children = [];
        if (parent) {
            parent.append(this);
        }
        else {
            this.parent = null;
        }
    }
    NestedSetNode.prototype.isLeaf = function () {
        return (this.depth > 0 && this.children.length === 0);
    };
    NestedSetNode.prototype.isRoot = function () {
        return !(this.parent);
    };
    NestedSetNode.prototype.prevSibling = function () {
        if (this.parent) {
            var idx = this.parent.children.indexOf(this);
            return (typeof this.parent.children[idx - 1] !== 'undefined') ? this.parent.children[idx - 1] : null;
        }
        else {
            return null;
        }
    };
    NestedSetNode.prototype.nextSibling = function () {
        if (this.parent) {
            var idx = this.parent.children.indexOf(this);
            return (typeof this.parent.children[idx + 1] !== 'undefined') ? this.parent.children[idx + 1] : null;
        }
        else {
            return null;
        }
    };
    NestedSetNode.prototype.countNextSiblings = function () {
        if (this.parent) {
            return this.parent.children.length - (this.parent.children.indexOf(this) + 1);
        }
        else {
            return 0;
        }
    };
    NestedSetNode.prototype.getSize = function () {
        if (this.isLeaf()) {
            return 2;
        }
        else {
            var childrenSize_1 = 0;
            this.children.forEach(function (child) {
                childrenSize_1 = childrenSize_1 + child.getSize();
            });
            return 2 + childrenSize_1;
        }
    };
    NestedSetNode.prototype.append = function (node) {
        node.parent = this;
        node.depth = this.depth + 1;
        this.children.push(node);
        this.rebuild();
    };
    NestedSetNode.prototype.prepend = function (node) {
        node.parent = this;
        node.depth = this.depth + 1;
        this.children.unshift(node);
        this.rebuild();
    };
    NestedSetNode.prototype.rebuild = function (list) {
        if (list === void 0) { list = []; }
        if (list.indexOf(this) === -1) {
            if (this.parent) {
                var myIdx = this.parent.children.indexOf(this);
                if (myIdx === 0) {
                    this.left = this.parent.left + 1;
                }
                else if (myIdx > 0) {
                    this.left = this.parent.children[myIdx - 1].right + 1;
                }
            }
            else {
                this.left = 0;
            }
            list.push(this);
        }
        if (!this.isLeaf()) {
            var tmp = [];
            tmp = this.children.reduce(function (accumulator, child) {
                if (accumulator.length === 0 && list.indexOf(child) === -1) {
                    accumulator.push(child);
                }
                return accumulator;
            }, tmp);
            if (tmp.length === 1) {
                this.right = this.left + this.getSize() - 1;
                return tmp[0].rebuild(list);
            }
        }
        if (this.isRoot()) {
            this.right = this.left + this.getSize() - 1;
            return list;
        }
        else {
            var sibling = this.nextSibling();
            this.right = this.left + this.getSize() - 1;
            return (sibling) ? sibling.rebuild(list) : this.parent.rebuild(list);
        }
    };
    NestedSetNode.prototype.toNestedSetProperties = function () {
        return {
            depth: this.depth,
            left: this.left,
            right: this.right,
            title: this.title
        };
    };
    NestedSetNode.prototype.flat = function () {
        return (this.rebuild()).map(function (node) { return node.toNestedSetProperties(); });
    };
    NestedSetNode.prototype.removeChild = function (child) {
        var idx = this.children.indexOf(child);
        if (idx > -1) {
            this.children.splice(idx, 1);
            child.parent = null;
            this.rebuild();
        }
    };
    NestedSetNode.prototype.validate = function () {
        if (this.depth !== 0 && this.parent === null) {
            throw new Error('Required parent when depth is not zero.');
        }
        if (this.parent !== null) {
            if (this.parent.left >= this.left) {
                throw new Error('Node left property cannot be lower than or equals to its parent left property.');
            }
            if (this.parent.right <= this.right) {
                throw new Error('Node right property cannot be greater than or equals to its parent right property.');
            }
            if (this.parent.depth + 1 !== this.depth) {
                throw new Error('Node depth property must be exactly plus-one of its parent depth property.');
            }
        }
    };
    return NestedSetNode;
}());
exports.NestedSetNode = NestedSetNode;
//# sourceMappingURL=nested-set-node.js.map